; ------------------------------------------------------------------------
; Commodore 64  Simple ANSI-Terminal Program v1.0
; ------------------------------------------------------------------------
;
; Copyright (c) 2008, Dennis Kuschel
; All rights reserved. 
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions
; are met:
;
;  1. Redistributions of source code must retain the above copyright
;     notice, this list of conditions and the following disclaimer.
;  2. Redistributions in binary form must reproduce the above copyright
;     notice, this list of conditions and the following disclaimer in the
;     documentation and/or other materials provided with the distribution.
;  3. The name of the author may not be used to endorse or promote
;     products derived from this software without specific prior written
;     permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
; INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
; (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
; STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
; OF THE POSSIBILITY OF SUCH DAMAGE.
;
; ------------------------------------------------------------------------
;
; This Program uses a UART "16C550" that is connected to the
; Expansion Port at I/O-Area 1 (signal -I/O1-, $DE00-$DEFF).
; It can be assembled as rom-able version or as disk-loadable prg file.
; The rom-able version can only be used in conjunction with the UART-
; cartridge that is described at www.mycpu.eu.
;
; To assemble this file you need the cc65 compiler from www.cc65.org.
;
; Use this lines to generate a .prg file:
;   ca65 -t c64 -D prgfile=1 terminal.s
;   ld65 -C c64prg.cfg terminal.o -o terminal.prg
;
; Use this lines to generate a rom image file:
;   ca65 -t c64 -D prgfile=0 terminal.s
;   ld65 -C c64rom.cfg terminal.o -o epromimage.bin
;
; If you have questions please contact me:
; mailto: dennis_k@freenet.de,  http://www.mycpu.eu
; ------------------------------------------------------------------------


;.define prgfile  0  ;set to 1 to generate a disk-loadable PRG file
.define swtest   0   ;set to 1 to enable software test mode
.define testcol  0
.define ansitest 0



; ------------------------------------------------------------------------
; Export Labels for debugging

    .export  processIncomingData
    .export  processOutgoingData
    .export  irqhandler
    .export  readUart
    .export  getFromRxBuf




; ------------------------------------------------------------------------
; Variables

;variables in Zero-Page
PTR1       = $03  ;2 bytes
PTR2       = $05  ;2 bytes
FIRSTLINE  = $16  ;1 byte
ADR16      = $17  ;2 bytes
TMP1       = $19  ;1 byte
TMP2       = $1A  ;1 byte
GFXADR     = $1B  ;2 byte   bitmap memory address (calculated from POSX+POSY+FIRSTLINE)
GFXNIB     = $1D  ;1 byte   character half: 0=left half, 1=right half (calculated from POSX)
GFXNMASK   = $1E  ;1 byte   character half: nibble mask
GFXCRTCODE = $1F  ;1 byte   character code for screen memory $0400-$07ff
GFXCRTADR  = $20  ;2 byte   address within screen memory $0400-$07ff
RXINPTR    = $22  ;2 bytes  receiver buffer: input pointer
RXOUTPTR   = $24  ;2 bytes  receiver buffer: output pointer
ESCSTATE   = $26  ;1 byte   escape-parser state
POSX       = $27  ;1 byte   current x-position
POSY       = $28  ;1 byte   current y-position
SAVEPOSX   = $29  ;1 byte   saved x-position
SAVEPOSY   = $2A  ;1 byte   saved y-position
CRSRSAVE   = $2B  ;2 byte   saved gfx bytes from current cursor position
ENABLECRSR = $2D  ;1 byte   flag: nonzero to enable the cursor
CRSRCLKDIV = $2E  ;1 byte   interrupt clock divider
REVERSE    = $2F  ;1 byte   flag: nonzero for reversed print
KEYSAMPLE  = $30  ;1 byte   flag: nonzero when the keyboard must be sampled
ESCNIDX    = $31  ;1 byte   escape number counter
PTR3       = $33  ;2 byte
NEXT_RIDX  = $35  ;1 byte
NEXT_RCTRL = $36  ;1 byte
FIRST_RIDX = $37  ;1 byte
MAX_RIDX   = $38  ;1 byte
NEXT_FRIDX = $39  ;1 byte
RXB_INPTR  = $3A  ;2 byte
RXB_OUTPTR = $3C  ;2 byte
TXB_INPTR  = $3E  ;1 byte
TXB_OUTPTR = $3F  ;1 byte
WRITETIMER = $40  ;1 byte   Timer for EEPROM writes
ENABLEIRQ  = $41  ;1 byte   flag: 0=UART-interrupt off, 1=UART-interrupt on
DISPMODE   = $42  ;1 byte   display mode: $80 = mono.text, $C0 = col.text, $00/$40 = hex
LASTPRINTED= $43  ;1 byte   last printed character
CURHEXROW  = $44  ;1 byte   current row for hex output (0..19)
FGCOLOR    = $45  ;1 byte   foreground color
BGCOLOR    = $46  ;1 byte   background color


;EEPROM configuration:
BAUDRATE   = $0338  ;1 byte   index for baudrate table
DATABITS   = $0339  ;1 byte   number of data bits: 7 or 8
STOPBITS   = $033A  ;1 byte   number of stop bits: 1 or 2
HANDSHAKE  = $033B  ;1 byte   type of handshake: 0=none, 1=hardware RTS/CTS
APPENDLF   = $033C  ;1 byte   flag: if nonzero append LF after CR when CR is received
CHARECHO   = $033D  ;1 byte   flag: if nonzero, echo all characters back to terminal
PARITY     = $033E  ;1 byte   type of parity: 0=none, 1=odd, 2=even
DISPLAYMODE= $033F  ;1 byte   display mode, copy of variable DISPMODE
MEMMAGIC   = $0340  ;1 byte   magic number: MAGICNUM
CHECKSUM   = $0341  ;1 byte   checksum for configuration data
CFG_START  = BAUDRATE
CFG_SIZE   = CHECKSUM - BAUDRATE + 1

;other variables
ESCNUMBERS = $0334  ;4 bytes  number in escape string
INITFLAG   = $0350  ;1 byte   nonzero when all is initialized
EEPROMERR  = $0351  ;1 byte   nonzero when the EEPROM could not be read
INTENSITY  = $0352  ;1 byte   nonzero when high intensity of foreground color
NEXTHCTOSND= $0353  ;1 byte   next handshake character to send (XON/XOFF)
LASTHCSENT = $0354  ;1 byte   last handshake character that was sent

;some defines
XON        = $11
XOFF       = $13
MAGICNUM   = $56
INITWRTIME = 75     ;3 seconds
MAXRASTERLN= 16
RASTERDIF  = 5
RASTERCNT  = 12
rasterl    = $03C0  ;2x14 bytes
cscmdtab   = $03E0  ;2x12 bytes

; --== Memory Mapping ==--
; Note: Code is at $8000-$BFFF, RAM at $D000-$FFFF is unused
SCRROUTINE2= $0800   ; $0800 - $1EFF  = $1700 bytes
TXBUFFER   = $1F00   ; $1F00 - $1FFF  = $0100 bytes
CHARGENMEM = $2000   ; $2000 - $3FFF  = $2000 bytes
SCRROUTINE1= $4000   ; $4000 - $56FF  = $1700 bytes
RXBFILLTAB = $5700   ; $5700 - $57FF  = $0100 bytes
RXBUFBASE  = $5800   ; $5800 - $7FFF  = $2800 bytes
CHARSETMEM = $C000   ; $C000 - $CFFF  = $1000 bytes
TXBUFSIZE  = $0100
MAX_RXBUF  = $28
CHARSETL   = CHARSETMEM
CHARSETR   = CHARSETMEM + $0800
UPPER_RXBUF= (RXBUFBASE / 256) + MAX_RXBUF


MONOFRAMECOL= 5
MONOBGCOL   = 0
MONOFGCOL   = 13

UARTBASE   = $DE28      ;any address with bits 0..2 = 0 in the range of $DE00-$DEFF
UART_RBR   = UARTBASE
UART_THR   = UARTBASE

.if prgfile
UART_IER   = $0200
.else
UART_IER   = UARTBASE+1
.endif

UART_IIR   = UARTBASE+2
UART_FCR   = UARTBASE+2
UART_LCR   = UARTBASE+3
UART_MCR   = UARTBASE+4
UART_LSR   = UARTBASE+5
UART_MSR   = UARTBASE+6
UART_SCR   = UARTBASE+7
UART_DLL   = UARTBASE
UART_DLM   = UARTBASE+1

I2CINREG   = UART_MSR
I2COUTREG  = UART_MCR
I2CIN_SCL  = $20  ;/DSR
I2CIN_SDA  = $80  ;/DCD
I2COUT_SCL = $04  ;OUT1
I2COUT_SDA = $08  ;OUT2
EE24C16ADR = $A0
MCRDEFAULT = $01


.data

;first visible raster line: 48   last raster line:  250

FIRSTRL   = 50
RASTER6L  = 48
MAXRLINE  = FIRSTRL + 4*RASTER6L


comprCharset:
.include "charset.s"

infotext: .byte  13,10,13,10,13,10
          .byte  "  Use the following keys to change the RS232-settings:",13,10,13,10
          .byte  "   C= + 1   change baudrate:  1200 - 115200 baud",13,10
          .byte  "   C= + 2   change databits:  5, 6, 7, 8",13,10
          .byte  "   C= + 3   change stopbits:  1 or 2",13,10
          .byte  "   C= + 4   change parity  :  none(off), odd or even parity",13,10
          .byte  "   C= + 5   change flow control:  none, hardware RTS/CTS, software XON/XOFF",13,10
          .byte  "   C= + 6   switch local echo on or off",13,10
          .byte  "   C= + 7   switch additional linefeed after carriage return on or off",13,10
          .byte  "   C= + H   toggle between text und hex mode",13,10
          .byte  "   C= + C   switch color mode on/off (experimental)",13,10
          .byte  13,10,13,10
          .byte  "  (C) 2007-2008 by Dennis Kuschel, Germany, WWW.MYCPU.EU",13,10,13,10,13,10,0
          
eepromErr: .byte " ERROR: Can not read configuration from EEPROM",13,10,13,10,0
eepWrErr: .byte 13," >> ERPROM write error << ",13,10,0

lineAdrl: .byte  $00,$40,$80,$C0,$00,$40,$00,$40,$80,$C0,$00,$40
          .byte  $00,$40,$80,$C0,$00,$40,$00,$40,$80,$C0,$00,$40
lineAdrh: .byte  $20,$21,$22,$23,$25,$26,$28,$29,$2A,$2B,$2D,$2E
          .byte  $30,$31,$32,$33,$35,$36,$38,$39,$3A,$3B,$3D,$3E
crtAdrl:  .byte  $28,$50,$78,$A0,$C8,$F0,$18,$40,$68,$90,$B8,$E0
          .byte  $08,$30,$58,$80,$A8,$D0,$F8,$20,$48,$70,$98,$C0
crtAdrh:  .byte  $04,$04,$04,$04,$04,$04,$05,$05,$05,$05,$05,$05
          .byte  $06,$06,$06,$06,$06,$06,$06,$07,$07,$07,$07,$07


linechr1: .byte  0,40,80,120,160,200,0,40,80,120,160,200
          .byte  0,40,80,120,160,200,0,40,80,120,160,200



rasttab:  ;array[24][4]

          .byte   58, 106, 154, 202
          .byte   98, 146, 194, 242
          .byte   90, 138, 186, 234
          .byte   82, 130, 178, 226
          .byte   74, 122, 170, 218
          .byte   66, 114, 162, 210

          .byte   58, 106, 154, 202
          .byte   98, 146, 194, 242
          .byte   90, 138, 186, 234
          .byte   82, 130, 178, 226
          .byte   74, 122, 170, 218
          .byte   66, 114, 162, 210

          .byte   58, 106, 154, 202
          .byte   98, 146, 194, 242
          .byte   90, 138, 186, 234
          .byte   82, 130, 178, 226
          .byte   74, 122, 170, 218
          .byte   66, 114, 162, 210

          .byte   58, 106, 154, 202
          .byte   98, 146, 194, 242
          .byte   90, 138, 186, 234
          .byte   82, 130, 178, 226
          .byte   74, 122, 170, 218
          .byte   66, 114, 162, 210

csettab:  ;array[24][4]
          .byte   24,  26,  28,  30
          .byte   26,  28,  30,  24
          .byte   26,  28,  30,  24
          .byte   26,  28,  30,  24
          .byte   26,  28,  30,  24
          .byte   26,  28,  30,  24

          .byte   26,  28,  30,  24
          .byte   28,  30,  24,  26
          .byte   28,  30,  24,  26
          .byte   28,  30,  24,  26
          .byte   28,  30,  24,  26
          .byte   28,  30,  24,  26

          .byte   28,  30,  24,  26
          .byte   30,  24,  26,  28
          .byte   30,  24,  26,  28
          .byte   30,  24,  26,  28
          .byte   30,  24,  26,  28
          .byte   30,  24,  26,  28

          .byte   30,  24,  26,  28
          .byte   24,  26,  28,  30
          .byte   24,  26,  28,  30
          .byte   24,  26,  28,  30
          .byte   24,  26,  28,  30
          .byte   24,  26,  28,  30


asciiColors:
asciiFgColCodes:    .byte  30,31,32,33,34,35,36,37,38,39
                    .byte  90,91,92,93,94,95,96,97,98,99
asciiBgColCodes:    .byte  40,41,42,43,44,45,46,47,48,49
                    .byte  100,101,102,103,104,105,106,107,108,109
petsciiColors:
petsciiFgColCodes:  .byte  0, 2, 5, 8, 6, 4, 11, 15, 15, 15
                    .byte  12, 10, 13, 7, 14, 4, 3, 1, 1, 1
petsciiBgColCodes:  .byte  0, 2, 5, 8, 6, 4, 11, 15, 0, 0
                    .byte  12, 10, 13, 7, 14, 4, 3, 1, 11, 11
COLTABSIZE = 40

headlineT:   .byte "ansi terminal v1.0                      ", 0
headlineH:   .byte "terminal ",27,"hex mode",29,"                     ", 0
hextab:      .byte $30,$31,$32,$33,$34,$35,$36,$37,$38,$39,$41,$42,$43,$44,$45,$46
brDivisors:  .byte    96,    48,    24,    12,      6,      3,      2,       1
baudRates :  .byte "1200",0,"2400",0,"4800",0,"9600",0,"19200",0,"38400",0,"57600",0,"115200",0
parities  :  .byte 'n','o','e'
handshakes:  .byte 'n','h','s'
MAXBAUDRATES = 7

;default UART configuration
DEF_BAUD   = 7  ;115200  4 ;19200 baud
DEF_DATA   = 8
DEF_STOP   = 1
DEF_HAND   = 1
DEF_LF     = 0
DEF_ECHO   = 0
DEF_PARITY = 0


.if ansitest
bufinitstart:
.include "bufinit.s"
bufinitend:
.endif



; ------------------------------------------------------------------------
; Start of CODE

.code
.segment       	"STARTUP"

.if prgfile

; BASIC header with a SYS call

        .word   $0801  ;Head            ; Load address
Head:   .word   $080b  ;@Next
        .word   0               ; Line number
        .byte   $9E,"2061"      ; SYS 2061
        .byte   $00             ; End of BASIC line
@Next:  .word   0               ; BASIC end marker
        clc
        bcc  cpprg

.endif

        .word   init
        .word   $FEBC
        .byte   $c3, $c2, $cd, $38, $30
init:
        stx  $D016
        jsr  $FDA3
        lda  #0
        tay
@loop:  sta  $0002,y
        sta  $0200,y
        sta  $0300,y
        iny
        bne  @loop
        ldx  #$FF
        ldy  #$7F
        jsr  $FD8C
        jsr  $FD15
        jsr  $FF5B
        cli


; ------------------------------------------------------------------------
; Cardridge initialization after hardware reset
; -> copies the EEPROM to RAM $7000-$FFFF ($D000-$DFFF remains free for I/O)

.if prgfile
        jmp  copyend
    ;-- move program to ROM-area $8000-
cpprg:  ldx  #$10
        ldy  #$08
        stx  PTR1
        sty  PTR1+1
        ldx  #$00
        ldy  #$80
        stx  PTR2
        sty  PTR2+1
        ldy  #0
@L1:    lda  (PTR1),y
        sta  (PTR2),y
        iny
        bne  @L1
        inc  PTR1+1
        inc  PTR2+1
        lda  PTR2+1
        cmp  #$cf
        bcc  @L1
        jmp  $fce2  ;reset        
.else
    ;-- copy from bank-switched EPROM to linear RAM --
        ;set destination pointer
        ldy  #$00
        ldx  #$80
        sty  PTR2
        stx  PTR2+1
        ;set page
        sty  TMP1
        ;copy the copy routine to RAM
@L5:    lda  @L0,y
        sta  $1000,y
        iny
        bne  @L5
        ;and call it
        jsr  $1000
        jmp  copyend
        ;copy routine
@L0:    inc  $d020
        ldy  #0
        ldx  #$80
        sty  PTR1
        stx  PTR1+1  ;source address
        ldx  #$20    ;blocks
@L1:    lda  (PTR1),y
        sta  (PTR2),y
        iny
        bne  @L1
        inc  PTR1+1
        inc  PTR2+1
        lda  PTR2+1
        cmp  #$D0    ;copy code from EPROM adr. $D000-$DFFF to $7000-$7FFF
        bne  @L2
        lda  #$70
@L2:    cmp  #$80
        bne  @L3
        lda  #$E0
@L3:    sta  PTR2+1
        dex
        bne  @L1
        ;next 8k block
        inc  TMP1
        lda  TMP1
        cmp  #4
        bcs  @L4
        asl
        asl
        sta  UART_MCR
        bne  @L0
@L4:    ;switch off the ROM and return
        lda  #MCRDEFAULT
        sta  UART_MCR
        rts
.endif
copyend:
        lda  #$36   ;switch off basic ROM
        sta  $01



; ------------------------------------------------------------------------
; Terminal code

        ;common initialization

        ;Screen
        lda  #0
        sta  INITFLAG
        lda  #$80
        sta  DISPMODE
        lda  #23
        sta  $D018
        jsr  setMonochromeMode
        lda  #0
        sta  NEXT_FRIDX
        sta  FIRST_RIDX
        sta  ENABLECRSR
        sta  CRSRCLKDIV
        sta  POSX
        sta  POSY
        sta  SAVEPOSX
        sta  SAVEPOSY
        sta  REVERSE
        sta  KEYSAMPLE
        sta  WRITETIMER
        sta  ENABLEIRQ
        sta  CURHEXROW
        sta  INTENSITY
        sta  NEXTHCTOSND
        sta  LASTHCSENT
        lda  #$80
        sta  DISPMODE
        sta  LASTPRINTED
        jsr  initUartBuffer
        jsr  clearCharset
        jsr  cursorHide
        jsr  generateCopyCode
        jsr  generateCharset
        jsr  initBell

        lda  #FIRSTRL-RASTERDIF-24
        sta  rasterl
        sta  rasterl+MAXRASTERLN
        lda  #FIRSTRL-24
        sta  rasterl+1
        sta  rasterl+MAXRASTERLN+1
        lda  #FIRSTRL+8-RASTERDIF
        sta  rasterl+2
        sta  rasterl+MAXRASTERLN+2
        lda  #FIRSTRL+8
        sta  rasterl+3
        sta  rasterl+MAXRASTERLN+3
        lda  #0
        sta  cscmdtab
        sta  cscmdtab+MAXRASTERLN
        lda  #23
        sta  cscmdtab+1
        sta  cscmdtab+MAXRASTERLN+1
        lda  #0
        jsr  setFirstLine
        lda  NEXT_FRIDX
        clc
        adc  #6
        sta  MAX_RIDX

        ;initialize the screen
        lda  #$FF
        ldy  #0
        clc
@L1:    sta  1064,y
        sta  1064+240,y
        sta  1064+480,y
        sta  1064+720,y
        iny
        cpy  #240
        bcc  @L1
        jsr  calcCharGfxAddr
        jsr  cursorEnable
        
        ;lock C= + Shift -function
        lda  $0291
        ora  #$80
        sta  $0291

        sei

        ;disable timer interrupt
        lda  #$7f
        sta  $dc0d

        ;set interrupt vector
        lda  #<irqhandler
        sta  $0314
        sta  $0316
        lda  #>irqhandler
        sta  $0315
        sta  $0317

        ;set first raster line
        ldx  NEXT_FRIDX
        stx  FIRST_RIDX
        lda  rasterl,x
        sta  $d012
        lda  $d011
        and  #$7f
        sta  $d011
        lda  cscmdtab,x
        sta  NEXT_RCTRL
        lda  cscmdtab,x
        sta  NEXT_RCTRL
        stx  NEXT_RIDX
   
        ;enable raster interrupt        
        lda  #$01
        sta  $d019
        sta  $d01a

        inc  INITFLAG  ;screen initialized now
        cli

        ;print help text
        ldx  #<infotext
        ldy  #>infotext
        jsr  printString

        ;initialize the UART by EEPROM configuration data
        jsr  i2cReadConfig
        bcc  @L2
        jsr  defaultConfig

        ldx  #<eepromErr
        ldy  #>eepromErr
        jsr  printString
        lda  #1
.if ansitest
        lda  #$C0
        sta  DISPLAYMODE
.else
        jmp  @L3
.endif

@L2:    lda  DISPLAYMODE
        sta  DISPMODE
        bit  DISPMODE
        bmi  @L4
        jsr  switchToHexMode
        jmp  @L5
@L4:    bvc  @L5
        jsr  setColorMode
@L5:    lda  #0

@L3:    sta  EEPROMERR
        jsr  setupUART
        jsr  printHeadline

.if ansitest
        jsr  processTestData
.endif

mainloop:   ; main program loop
        jsr  mainServices
        jmp  mainloop   
      


mainServices:   ;must be called periodically

.if swtest
    ;simulate a UART interrupt
    lda  UART_IER
    and  #1
    beq  @noi
;    brk
    nop
    nop
@noi:
.endif

        jsr  uartoutput
        jsr  getFromRxBuf
        beq  @ML02
        jsr  processIncomingData
        jsr  getFromRxBuf
        beq  @ML02
        jsr  processIncomingData
        jmp  @ML03
@ML02:  jsr  handshakeON  ;buffer empty, turn transmitter on
        jsr  doCursor     ;let the cursor blink
@ML03:  lda  KEYSAMPLE
        cmp  #2
        bcc  @ML00
        lda  #0
        sta  KEYSAMPLE
        jsr  checkEEPROMwrite  ;check if data must be written to EEPROM
        jsr  scanKeyboard  ;scan keyboard. This takes really long!
        jsr  readkey
        beq  @ML00
        cmp  #129
        bne  @ML04
        jmp  changeBaudrate
@ML04:  cmp  #149
        bne  @ML05
        jmp  changeDatabits
@ML05:  cmp  #150
        bne  @ML06
        jmp  changeStopbits
@ML06:  cmp  #151
        bne  @ML07
        jmp  changeParity
@ML07:  cmp  #152
        bne  @ML08
        jmp  changeHandshake
@ML08:  cmp  #153
        bne  @ML09
        jmp  changeEcho
@ML09:  cmp  #154
        bne  @ML10
        jmp  changeLinefeed
@ML10:  cmp  #180   ;C= + 'h'
        bne  @ML11
        jmp  changePrintMode
@ML11:  cmp  #188   ;C= + 'c'
        bne  @ML12
        jmp  changeColorMode
@ML12:  jsr  processOutgoingData
@ML00:  rts


changeBaudrate:
        ldx  BAUDRATE
        inx
        cpx  #MAXBAUDRATES+1
        bcc  @L1
        ldx  #0
@L1:    stx  BAUDRATE
        jsr  setupUART
        jsr  printHeadline
        jmp  saveConfig

changeDatabits:
        ldx  DATABITS
        inx
        cpx  #9
        bcc  @L1
        ldx  #5
@L1:    stx  DATABITS
        jsr  setupUART
        jsr  printHeadline
        jmp  saveConfig

changeStopbits:
        ldx  STOPBITS
        inx
        cpx  #3
        bcc  @L1
        ldx  #1
@L1:    stx  STOPBITS
        jsr  setupUART
        jsr  printHeadline
        jmp  saveConfig

changeParity:
        ldx  PARITY
        inx
        cpx  #3
        bcc  @L1
        ldx  #0
@L1:    stx  PARITY
        jsr  setupUART
        jsr  printHeadline
        jmp  saveConfig

changeHandshake:
        ldx  HANDSHAKE
        inx
        cpx  #3
        bcc  @L1
        ldx  #0
@L1:    stx  HANDSHAKE
        jsr  setupUART
        jsr  printHeadline
        jmp  saveConfig

changeEcho:
        lda  CHARECHO
        eor  #1
        sta  CHARECHO
        jsr  setupUART
        jsr  printHeadline
        jmp  saveConfig

changeLinefeed:
        lda  APPENDLF
        eor  #1
        sta  APPENDLF
        jsr  setupUART
        jsr  printHeadline
        jmp  saveConfig

changeColorMode:
        bit  DISPMODE
        bpl  @L1   ;don't toggle color mode when we are in hex mode
        bvc  @L0
        jsr  setMonochromeMode
        jmp  saveConfig
@L0:    jsr  setColorMode
        jmp  saveConfig
@L1:    rts

changePrintMode:
        lda  DISPMODE
        bpl  @L1
        jsr  switchToHexMode
        jmp  saveConfig
@L1:    jsr  switchToTextMode
        jmp  saveConfig

saveConfig:
        lda  DISPMODE
        sta  DISPLAYMODE
        lda  #INITWRTIME
        sta  WRITETIMER
        rts


readkey:
        ;warning! it is NOT allowed to disable interrupts in the main prog!
        lda  $c6
        beq  nokey
        ldx  #0
        lda  $0277
        stx  $c6
nokey:  rts


printString:
        stx  PTR3
        sty  PTR3+1
@LOOP:  ldy  #0
        lda  (PTR3),y
        beq  @L0
        cmp  #65
        bcc  @L1
        cmp  #91
        bcc  @L2
        cmp  #97
        bcc  @L1
        cmp  #123
        bcs  @L1
@L2:    eor  #$20
@L1:    and  #$7F
        jsr  putch
        inc  PTR3
        bne  @LOOP
        inc  PTR3+1
        bne  @LOOP
@L0:    rts




; ------------------------------------------------------------------------
; Process incoming data, parse VT100 sequences, and display

dispHex:
        jmp  outputHex

processIncomingData:    ;data in Accu
        ldx  DISPMODE
        bpl  dispHex
        ldx  ESCSTATE     ;0=no esc, 1=esc received, 2='[' received,
        bne  @L2
        cmp  #32
        bcs  @L38
        cmp  #8
        beq  @L34
        cmp  #9
        beq  @L35
        cmp  #7
        beq  @bell
        cmp  #27          ;ESC
        beq  @L1          ;state := 1
        cmp  #13
        bne  @L38
        ldx  APPENDLF
        beq  @L38
        jsr  putch
        lda  #10
@L38:   jmp  putch
@L1:    ldx  #0
        stx  ESCNUMBERS
        stx  ESCNIDX
        inx
        stx  ESCSTATE
        rts
@L3:    inc  ESCSTATE
        rts

@bell:  jmp  bell

@L34:   ;backspace
        lda  #1
        jmp  @L36

@L35:   ;tab
        lda  POSX
        clc
        adc  #7
        and  #$F8
        cmp  #80
        bcc  @L37
        lda  #79
@L37:   tax
        ldy  POSY
        jmp  gotoxy

@L2:    ;parse ESC sequence
        cmp  #27
        beq  @L1  ;esc in esc-sequence, start new sequence
        dex
        bne  @L4          ;jump if state != 1
        cmp  #$5B  ;'['
        beq  @L3          ;state := 2
        stx  ESCSTATE     ;state := 0
        jmp  putch
@L4:    dex
        bne  @L5          ;jump if state != 2
        ;check for number
        cmp  #$30
        bcc  @L7
        cmp  #$3A
        bcs  @L7
        ;NUMBER *10 + new
        pha
        ldy  ESCNIDX
        lda  ESCNUMBERS,y
        asl
        sta  ESCNUMBERS,y
        asl
        asl
        adc  ESCNUMBERS,y
        sta  ESCNUMBERS,y
        pla
        and  #$0F
        clc
        adc  ESCNUMBERS,y
        sta  ESCNUMBERS,y
        rts  ;done
@L7:    ;no more numbers, is a semi-colon?
        inc  ESCNIDX
        cmp  #59
        bne  @L9
        ;yes, read more numbers
        ldy  ESCNIDX
        cpy  #4
        bcs  @L3         ;state := 3
        lda  #0
        sta  ESCNUMBERS,y
        rts
@L9:    inc  ESCSTATE    ;state := 3
        inx
@L5:    dex
        beq  @L8         ;jump if state == 3
        ;unknown command
        ldx  #0
        stx  ESCSTATE
@L12:   pha
        lda  #$5B  ;'['
        jsr  putch
        pla
        jmp  putch
@L8:    ;reset state machine
        stx  ESCSTATE
        ;convert small letters to big ones
        cmp  #97
        bcc  @L10
        cmp  #123
        bcs  @L10
        and  #$DF
@L10:   ;parse the ESC-command
        ldy  ESCNIDX
        cmp  #72  ;'H'
        bne  @L11
        ;set cursor position X/Y
@L27:   cpy  #2
        bcs  @L33
        lda  #0
        sta  ESCNUMBERS+1
@L33:   lda  ESCNUMBERS+0
        jsr  zero2one
        tay
        lda  ESCNUMBERS+1
        jsr  zero2one
        tax
        dex
        dey
        jmp  gotoxy
@L11:   cmp  #70  ;'F'
        beq  @L27
        cmp  #65  ;'A'
        bne  @L13
        ;move cursor up
        lda  ESCNUMBERS+0
        jsr  zero2one
        sta  TMP1
        lda  POSY
        sec
        sbc  TMP1
        bcs  @L15
        lda  #0
@L15:   jmp  gotoy
@L13:   cmp  #66  ;'B'
        bne  @L14
        ;move cursor down
        lda  ESCNUMBERS+0
        jsr  zero2one
@L16:   pha
        jsr  printLF
        pla
        tax
        dex
        txa
        bne  @L16
        rts
@L14:   cmp  #67  ;'C'
        bne  @L17
        ;move cursor right
        lda  ESCNUMBERS+0
        jsr  zero2one
        clc
        adc  POSX
        cmp  #79
        bcc  @L18
        lda  #79
@L18:   jmp  gotox
@L17:   cmp  #68  ;'D'
        bne  @L19
        ;move cursor left
        lda  ESCNUMBERS+0
        jsr  zero2one
@L36:   sta  TMP1
        lda  POSX
        sec
        sbc  TMP1
        bcs  @L20
        lda  #0
@L20:   jmp  gotox
@L19:   cmp  #78  ;'N'
        bne  @L21
        lda  ESCNUMBERS+0
        cmp  #6
        bne  @L23
        ;ask for cursor position, send cursor pos. to host
        lda  #27
        jsr  putToTxBuf
        lda  #$5B ;'['
        jsr  putToTxBuf
        ldy  POSY
        iny
        tya
        jsr  sendNumber
        lda  #59  ;semi-colon
        jsr  putToTxBuf
        ldy  POSX
        iny
        tya
        jsr  sendNumber
        lda  #82  ;'R'
        jmp  putToTxBuf
@L21:   cmp  #83  ;'S'
        bne  @L22
        ;save cursor position
        lda  POSX
        sta  SAVEPOSX
        lda  POSY
        sta  SAVEPOSY
@L23:   rts
@L22:   cmp  #85  ;'U'
        bne  @L24
        ;restore cursor position
        ldx  SAVEPOSX
        ldy  SAVEPOSY
        jmp  gotoxy
@L24:   cmp  #74  ;'J'
        bne  @L25
        ;clear the screen
        lda  ESCNUMBERS+0
        cmp  #3
        bcs  @L23
        jmp  clearScreen
@L25:   cmp  #75  ;'K'
        bne  @L26
        ;clear the line from cusor X-Pos
        jmp  clearLine
@L26:   cmp  #77  ;'M'
        bne  @L28
        ;change modus
        ldy  ESCNIDX
        beq  @L29
        ldy  #0
@L30:   lda  ESCNUMBERS,y
        bne  @L31
        ;todo: set other defaults here
        lda  #0
        sta  BGCOLOR
        sta  INTENSITY
        lda  #15
        sta  FGCOLOR
@L40:   lda  #0
        sta  REVERSE
        beq  @L32
@L31:   cmp  #3
        bcs  @L44
        jsr  intensFgColor
        lda  #1
        sta  INTENSITY   ;set/reset high intensity or bold
        bne  @L32
@L44:   cmp  #7
        bne  @L39
        sta  REVERSE
        beq  @L32
@L39:   cmp  #27
        beq  @L40
        ;check color codes
        ldx  #COLTABSIZE-1
@L41:   cmp  asciiColors,x
        beq  @L42
        dex
        bpl  @L41
@L32:   iny
        cpy  ESCNIDX
        bcc  @L30
@L29:   rts
@L42:   lda  INTENSITY
        beq  @L45  ;jump if not high intensity color
        cpx  #COLTABSIZE/4
        bcs  @L45
        txa
        clc
        adc  #COLTABSIZE/4
        tax
@L45:   lda  petsciiColors,x
        cpx  #COLTABSIZE/2
        bcs  @L43  ;jump if background color
        sta  FGCOLOR
        bcc  @L32
@L43:   sta  BGCOLOR
        bcs  @L32
@L28:   ;unknown command
        jmp  @L12

intensFgColor:   ;increase intensity of foreground color
        lda  FGCOLOR
        ldx  #0
@L1:    cmp  petsciiFgColCodes,x
        beq  @L2
        inx
        cpx  #10
        bcc  @L1
        rts
@L2:    lda  petsciiFgColCodes+10,x
        sta  FGCOLOR
        rts


zero2one:
        cmp  #0
        bne  @L0
        lda  #1
@L0:    rts

sendNumber:
        cmp  #10
        bcc  @L0
        ldx  #0
@L2:    cmp  #10
        bcc  @L1
        sbc  #10
        inx
        bne  @L2
@L1:    pha
        txa
        ora  #$30
        jsr  putToTxBuf
        pla
@L0:    ora  #$30
        jmp  putToTxBuf


initBell:
        lda  #0
        sta  $D404
        sta  $D405
        sta  $D406
        lda  #15     ;Lautstrke
        sta  $D418
        rts

bell:   ;output a bell sound
        pha
        jsr  initBell
        lda  #$06
        sta  $D405
        lda  #$0F
        sta  $D406
        ldx  #193    ;Frequenz
        ldy  #73 
        stx  $D400
        sty  $D401
        lda  #17     ;Wellenform
        sta  $D404
        pla
        rts



; ------------------------------------------------------------------------
; Process outgoing data, construct VT100 sequences


processOutgoingData:    ;call with Accu = key code
        pha
        jsr  hasTxbufSpace
        pla
        bcs  @L25 ;buffer full
        cmp  #32
        bcc  @L2
        cmp  #128
        bcc  @L1
        cmp  #160
        bcc  @L2
        
        cmp  #164  ;underline '_'
        beq  @L27
        cmp  #175
        bne  @L26
@L27:   lda  #95   ;_
        bne  @L1

@L26:   cmp  #222  ;pi -> tilde
        bne  @L1
        lda  #126

@L1:    ;send usual characters
        jsr  petscii2ascii
        jmp  sendout
@L25:   rts

@L2:    ;send special characters
        cmp  #17
        bne  @L4
        lda  #66  ;'B' -> crsr down
        jmp  sendShortEsc
@L4:    cmp  #29
        bne  @L5
        lda  #67  ;'C' -> crsr right
        jmp  sendShortEsc
@L5:    cmp  #145
        bne  @L6
        lda  #65  ;'A' -> crsr up
        jmp  sendShortEsc
@L6:    cmp  #157
        bne  @L7
        lda  #68  ;'D' -> crsr left
        jmp  sendShortEsc
@L7:    cmp  #19
        bne  @L8
        lda  #72  ;'H' -> crsr home
        jmp  sendShortEsc
@L8:    cmp  #20
        bne  @L9
        lda  #8         ;backspace
        jmp  sendout
@L9:    cmp  #18
        bne  @L10
        jsr  startEsc   ;send reverse-on
        lda  #$37
@L13:   jsr  sendout
        lda  #109  ;'m'
        jmp  sendout
@L10:   cmp  #141
        bne  @L11
        lda  #13
        jmp  sendout    ;send return
@L11:   cmp  #146
        bne  @L12
        jsr  startEsc   ;send reverse-off
        lda  #$30
        jmp  @L13
@L12:   cmp  #147
        bne  @L14
        ;This is "clear screen". We clear the _local_ screen.
        lda  #0
        sta  LASTPRINTED
        jmp  clearScreen
        ;jsr  startEsc   ;send clear-screen
        ;lda  #$32
        ;jsr  sendout
        ;lda  #74  ;'J'
        ;jmp  sendout
@L14:   cmp  #148       ;insert-key
        bne  @L15
        lda  #27        ;send ESC when insert-key is pressed
        jmp  sendout
@L15:   cmp  #133
        bne  @L17
        lda  #80        ;F1
        jmp  sendFKey
@L17:   cmp  #134
        bne  @L18
        lda  #82        ;F3
        jmp  sendFKey
@L18:   cmp  #135
        bne  @L19
        lda  #84        ;F5
        jmp  sendFKey
@L19:   cmp  #136
        bne  @L20
        lda  #86        ;F7
        jmp  sendFKey
@L20:   cmp  #137
        bne  @L21
        lda  #81        ;F2
        jmp  sendFKey
@L21:   cmp  #138
        bne  @L22
        lda  #83        ;F4
        jmp  sendFKey
@L22:   cmp  #139
        bne  @L23
        lda  #85        ;F6
        jmp  sendFKey
@L23:   cmp  #140
        bne  @L16
        lda  #87        ;F8
        jmp  sendFKey
@L16:   cmp  #32
        bcs  @L0
        jmp  sendout    ;send anything else, e.g. CTRL+C
@L0:    rts

startEsc:
        lda  #27
        jsr  sendout
        lda  #$5B  ;'['
        jmp  sendout

sendFKey:
        pha
        lda  #27
        jsr  sendout
        lda  #79
        jsr  sendout
        pla
        jmp  sendout

sendShortEsc:
        pha
        jsr  startEsc
        pla
        jmp  sendout

sendout:
        ldx  CHARECHO
        beq  @L1
        pha
        jsr  processIncomingData
        pla
@L1:    jmp  putToTxBuf




; ------------------------------------------------------------------------
; IRQ handler


readUart:  ;read max. 5 byte from the UART   -> 40 + 9 + 9 + 5*21 + 7 + 22 = 192
        ldx  RXB_INPTR+1    ;3
        lda  RXBFILLTAB,x   ;4
        tay                 ;2
        lda  UART_LSR       ;4
.if swtest
        sec                 ;2
.else
        lsr                 ;2
.endif
        bcc  iret           ;3
        lda  UART_RBR       ;4
        sta  (RXB_INPTR),y  ;6
        iny                 ;2   -> sum 21
        lda  UART_LSR       ;4
.if swtest
        sec                 ;2
.else
        lsr                 ;2
.endif
        bcc  rxrdy          ;3
        lda  UART_RBR       ;4
        sta  (RXB_INPTR),y  ;6
        iny                 ;2
        lda  UART_LSR       ;4
.if swtest
        sec                 ;2
.else
        lsr                 ;2
.endif
        bcc  rxrdy          ;3
        lda  UART_RBR       ;4
        sta  (RXB_INPTR),y  ;6
        iny                 ;2
        lda  UART_LSR       ;4
.if swtest
        sec                 ;2
.else
        lsr                 ;2
.endif
        bcc  rxrdy          ;3
        lda  UART_RBR       ;4
        sta  (RXB_INPTR),y  ;6
        iny                 ;2
rxloop: lda  UART_LSR       ;4
        lsr                 ;2
        bcc  rxrdy          ;3
morerx:
        lda  UART_RBR       ;4
        sta  (RXB_INPTR),y  ;6
        iny                 ;2
rxrdy:  tya                 ;2
        sta  RXBFILLTAB,x   ;5
iret:   pla                 ;4
        tay                 ;2
        pla                 ;4
        tax                 ;2
        pla                 ;4
        rti                 ;6


serIntOff:
        lda  #0
        sta  UART_IER
        jmp  nextrint


irqhandler:
        lda  $d019  ;interrupt pending register
        and  #1
        beq  readUart  ; no raster interrupt, assume UART as source: read bytes from UART
        sta  $d019  ;clear the interrupt

        lda  NEXT_RCTRL
.if testcol
        sta  $d020        ;-- demo
.endif
        beq  serIntOff
        sta  $d018   ;VIC memory base address register
        lda  ENABLEIRQ
        sta  UART_IER  ;enable interrupts

nextrint:
        ldx  NEXT_RIDX
        inx
        ;set next raster line number for interrupt
        cpx  MAX_RIDX
        bcc  @L0

        ;--> inserted: do some services
        ;swap UART receiver buffer
        ldx  RXB_INPTR+1
        lda  RXBFILLTAB,x
        beq  @L3
        inx
        cpx  #UPPER_RXBUF
        bcc  @L4
        ldx  #>RXBUFBASE
@L4:    stx  RXB_INPTR+1
        lda  #0
        sta  RXBFILLTAB,x
        ;flow control: disable transmitter if required
        ldy  HANDSHAKE
        beq  @L3
        txa
        sec
        sbc  RXB_OUTPTR+1
        bcs  @L5
        adc  #MAX_RXBUF
@L5:    cmp  #MAX_RXBUF/2
        bcc  @L3
        cpy  #1
        beq  @L6
        ;do software handshake
        lda  #XOFF
        sta  NEXTHCTOSND
        jmp  @L3
@L6:    ;do hardware handshake
        lda  UART_MCR
        and  #$FD       ;set /RTS to high-level
        sta  UART_MCR
        
@L3:    ;set flag: keyboard must be sampled
        inc  KEYSAMPLE
        ;decrement cursor blink counter
        lda  CRSRCLKDIV
        beq  @L2
        dec  CRSRCLKDIV
@L2:    ;<-- end of services

        lda  NEXT_FRIDX
        sta  FIRST_RIDX
        tax
        clc
        adc  #RASTERCNT
        sta  MAX_RIDX

@L0:    lda  rasterl,x
        cmp  rasterl+2,x
        bne  @L1
        inx
        inx
        lda  rasterl,x
@L1:    sta  $d012

        lda  cscmdtab,x
        sta  NEXT_RCTRL
        stx  NEXT_RIDX
        bne  norast

        lda  UART_LSR       ;4
.if swtest
        clc
.else
        lsr                 ;2
.endif
        bcs  isrreaduart

norast: ;return from interrupt
        pla
        tay
        pla
        tax
        pla
        rti

isrreaduart:
        bcc  norast         ;3
        ldx  RXB_INPTR+1    ;3
        lda  RXBFILLTAB,x   ;4
        tay                 ;2
        jmp  morerx



; ------------------------------------------------------------------------
; NMI handler

nmihandler:
        rti




; ------------------------------------------------------------------------
; Generate character sets

generateCharset:
        lda  #<comprCharset
        sta  PTR1
        lda  #>comprCharset
        sta  PTR1+1
        lda  #<CHARSETR
        sta  PTR2
        lda  #>CHARSETR
        sta  PTR2+1

        ;clear first 32 characters
        ldy  #0
        tya
@L1:    sta  (PTR2),y
        iny
        bne  @L1
        inc  PTR2+1

        ;copy 48 characters, low-nibble
        lda  #48
        sta  TMP1
@L2:    ldx  #8
@L3:    lda  (PTR1),y
        and  #$0F
        sta  (PTR2),y
        iny
        bne  @L4
        inc  PTR1+1
        inc  PTR2+1
@L4:    dex
        bne  @L3
        dec  TMP1
        bne  @L2

        ;copy 48 characters, high-nibble
        tya
        clc
        adc  PTR2
        sta  PTR2
        ldy  #0
        lda  #>comprCharset
        sta  PTR1+1
        lda  #48
        sta  TMP1
@L5:    ldx  #8
@L6:    lda  (PTR1),y
        lsr
        lsr
        lsr
        lsr
        and  #$0F
        sta  (PTR2),y
        iny
        bne  @L7
        inc  PTR1+1
        inc  PTR2+1
@L7:    dex
        bne  @L6
        dec  TMP1
        bne  @L5

        ;create inverted version of first 128 characters
        lda  #<CHARSETR
        sta  PTR1
        lda  #>CHARSETR
        sta  PTR1+1
        lda  #<(CHARSETR + $0400)
        sta  PTR2
        lda  #>(CHARSETR + $0400)
        sta  PTR2+1
        ldy  #0
        lda  #4
        sta  TMP1
@L8:    lda  (PTR1),y
        eor  #$0F
        sta  (PTR2),y
        iny
        bne  @L8
        inc  PTR1+1
        inc  PTR2+1
        dec  TMP1
        bne  @L8

        ;create swaped version of the first character set
        lda  #<CHARSETR
        sta  PTR1
        lda  #>CHARSETR
        sta  PTR1+1
        lda  #<CHARSETL
        sta  PTR2
        lda  #>CHARSETL
        sta  PTR2+1
        lda  #8
        sta  TMP1
@L9:    lda  (PTR1),y
        asl
        asl
        asl
        asl
        sta  (PTR2),y
        iny
        bne  @L9
        inc  PTR1+1
        inc  PTR2+1
        dec  TMP1
        bne  @L9
        
        ;all done.
        rts


clearCharset:   ;clear character code $ff in charset bitmaps
        ldy  #7
        lda  #0
@L1:    sta  CHARGENMEM+$07F8,y
        sta  CHARGENMEM+$0FF8,y
        sta  CHARGENMEM+$17F8,y
        sta  CHARGENMEM+$1FF8,y
        dey
        bpl  @L1
        rts




; ------------------------------------------------------------------------
; Helper functions


setFirstLine:
        sta  FIRSTLINE
        asl
        asl
        tay
        lda  NEXT_FRIDX
        eor  #MAXRASTERLN
        and  #MAXRASTERLN
        tax

        lda  rasttab,y
        sta  rasterl+5,x
        sec
        sbc  #RASTERDIF
        sta  rasterl+4,x
        lda  csettab,y
        sta  cscmdtab+5,x
        lda  #0
        sta  cscmdtab+4,x

        lda  rasttab+1,y
        sta  rasterl+7,x
        sec
        sbc  #RASTERDIF
        sta  rasterl+6,x
        lda  csettab+1,y
        sta  cscmdtab+7,x
        lda  #0
        sta  cscmdtab+6,x

        lda  rasttab+2,y
        sta  rasterl+9,x
        sec
        sbc  #RASTERDIF
        sta  rasterl+8,x
        lda  csettab+2,y
        sta  cscmdtab+9,x
        lda  #0
        sta  cscmdtab+8,x

        lda  rasttab+3,y
        sta  rasterl+11,x
        sec
        sbc  #RASTERDIF
        sta  rasterl+10,x
        lda  csettab+3,y
        sta  cscmdtab+11,x
        sta  cscmdtab+3,x
        lda  #0
        sta  cscmdtab+10,x
        sta  cscmdtab+2,x

        stx  NEXT_FRIDX
        rts


generateCopyCode:
        ;generate screen character scroll up routine
        lda  #<SCRROUTINE1
        sta  PTR1
        lda  #>SCRROUTINE1
        sta  PTR1+1
        
        lda  #<1064
        sta  ADR16
        lda  #>1064
        sta  ADR16+1

@L1:    ldy  #0

        ;"LDA PTR1+40"
        lda  #$AD
        sta  (PTR1),y
        iny
        clc
        lda  ADR16
        adc  #40
        sta  (PTR1),y
        iny
        lda  ADR16+1
        adc  #0
        sta  (PTR1),y
        iny
        ;"STA PTR1"
        lda  #$8D
        sta  (PTR1),y
        iny
        lda  ADR16
        sta  (PTR1),y
        iny
        lda  ADR16+1
        sta  (PTR1),y
        iny
        
        clc
        lda  PTR1
        adc  #6
        sta  PTR1
        lda  PTR1+1
        adc  #0
        sta  PTR1+1
        
        inc  ADR16
        bne  @L2
        inc  ADR16+1
@L2:    lda  ADR16
        cmp  #<1984
        bne  @L1
        lda  ADR16+1
        cmp  #>1984
        bne  @L1
        
        ;"LDA #$FF"
        ldy  #0
        lda  #$A9
        jsr  @putb
        lda  #$FF
        jsr  @putb
        
        ;40x "STA $07xx"
        ldx  #40
@L3:    lda  #$8D
        jsr  @putb
        lda  ADR16
        jsr  @putb
        lda  ADR16+1
        jsr  @putb
        inc  ADR16
        dex
        bne  @L3
        
        ;add "RTS"
        lda  #$60
        jsr  @putb

        ;generate screen color scroll up routine
        lda  #<SCRROUTINE2
        sta  PTR1
        lda  #>SCRROUTINE2
        sta  PTR1+1
        
        lda  #<55336
        sta  ADR16
        lda  #>55336
        sta  ADR16+1

@L5:    ldy  #0

        ;"LDA PTR1+40"
        lda  #$AD
        sta  (PTR1),y
        iny
        clc
        lda  ADR16
        adc  #40
        sta  (PTR1),y
        iny
        lda  ADR16+1
        adc  #0
        sta  (PTR1),y
        iny
        ;"STA PTR1"
        lda  #$8D
        sta  (PTR1),y
        iny
        lda  ADR16
        sta  (PTR1),y
        iny
        lda  ADR16+1
        sta  (PTR1),y
        iny
        
        clc
        lda  PTR1
        adc  #6
        sta  PTR1
        lda  PTR1+1
        adc  #0
        sta  PTR1+1
        
        inc  ADR16
        bne  @L6
        inc  ADR16+1
@L6:    lda  ADR16
        cmp  #<56256
        bne  @L5
        lda  ADR16+1
        cmp  #>56256
        bne  @L5
        
        ;"LDA FGCOLOR"
        ldy  #0
        lda  #$A5
        jsr  @putb
        lda  #FGCOLOR
        jsr  @putb
        
        ;40x "STA $DBxx"
        ldx  #40
@L7:    lda  #$8D
        jsr  @putb
        lda  ADR16
        jsr  @putb
        lda  ADR16+1
        jsr  @putb
        inc  ADR16
        dex
        bne  @L7
        
        ;add "RTS"
        lda  #$60
@putb:  sta  (PTR1),y
        inc  PTR1
        bne  @L4
        inc  PTR1+1
@L4:    rts



calcCharGfxAddr:
        ;calculate character graphic address in memory
        ;in : FIRSTLINE, POSX, POSY
        ;out: GFXADR, GFXNIB, GFXNMASK, GFXCRTCODE, GFXCRTADR

        ;Y = (FIRSTLINE + POSY) % 24
        lda  FIRSTLINE
        clc
        adc  POSY
        cmp  #24
        bcc  @L1
        sbc  #24
@L1:    tay
calcCGA:

        ;A = (POSX/2)*8
        lda  POSX
        and  #$FE
        asl
        asl
        php
        ;P = lineAdr,Y + A
        clc
        adc  lineAdrl,y
        sta  GFXADR
        lda  #0
        adc  lineAdrh,y
        plp
        adc  #0
        sta  GFXADR+1
        ;get character side (left or right)
        lda  POSX
        ror
        bcs  @L2
        lda  #0
        sta  GFXNIB
        lda  #$0F
        jmp  @L3
@L2:    lda  #8
        sta  GFXNIB
        lda  #$F0
@L3:    sta  GFXNMASK

        ;check code, clear memory if neccessary
        ldx  POSY
        lda  POSX
        lsr
        clc
        adc  crtAdrl,x
        sta  GFXCRTADR
        lda  #0
        adc  crtAdrh,x
        sta  GFXCRTADR+1

        lda  POSX
        lsr
        clc
        adc  linechr1,y
        sta  GFXCRTCODE
        rts



outputCharacter:
        ;ASCII character code in accu
        tax
        bpl  @L5
        lda  #'0'
@L5:    ldy  REVERSE
        beq  @L0
        ora  #$80
@L0:    sta  PTR1

        ;set color if color mode is enabled
        bit  DISPMODE
        bvc  @L4
        bpl  @L4
        cmp  #$21
        bcc  @L4
        lda  GFXCRTADR
        sta  @CPTR+1        ;this is self-modifying code
        lda  GFXCRTADR+1
        adc  #$D4-1
        sta  @CPTR+2
        lda  FGCOLOR
@CPTR:  sta  $D800
        lda  BGCOLOR
        sta  $D021
@L4:    
        ldx  GFXNMASK
        ldy  #0
        sty  ENABLECRSR      ;cursor off
        lda  (GFXCRTADR),y
        cmp  GFXCRTCODE
        beq  @L1
        ;clear bitmap
        ldx  #0
@L1:    stx  TMP1
        ;calculate the character bitmap source address
        ;PTR1 = PTR1*8 + CHARSETx + 256*GFXNIB
        tya
        asl  PTR1
        rol
        asl  PTR1
        rol
        asl  PTR1
        rol
        clc
        adc  #>CHARSETMEM
        adc  GFXNIB
        sta  PTR1+1

        ;set the character (fast, unrolled loop)
        ldy  #7
        lda  (GFXADR),y
        and  TMP1
        ora  (PTR1),y
        sta  (GFXADR),y
        dey
        lda  (GFXADR),y
        and  TMP1
        ora  (PTR1),y
        sta  (GFXADR),y
        dey
        lda  (GFXADR),y
        and  TMP1
        ora  (PTR1),y
        sta  (GFXADR),y
        dey
        lda  (GFXADR),y
        and  TMP1
        ora  (PTR1),y
        sta  (GFXADR),y
        dey
        lda  (GFXADR),y
        and  TMP1
        ora  (PTR1),y
        sta  (GFXADR),y
        dey
        lda  (GFXADR),y
        and  TMP1
        ora  (PTR1),y
        sta  (GFXADR),y
        dey
        lda  (GFXADR),y
        and  TMP1
        ora  (PTR1),y
        sta  (GFXADR),y
        dey
        lda  (GFXADR),y
        and  TMP1
        ora  (PTR1),y
        sta  (GFXADR),y

        ;set ctrl character
        lda  GFXCRTCODE
        sta  (GFXCRTADR),y
        rts


petscii2ascii:
        cmp  #65
        bcc  @L0
        cmp  #91
        bcs  @L1
        adc  #32
@L0:    rts
@L1:    cmp  #193
        bcc  @L0
        cmp  #219
        bcs  @L2
        sec
        sbc  #128
        rts
@L2:    and  #$7F
        jmp  petscii2ascii




;------------------------------------------------------------------------------
; Cursor / Screen Functions


scrollUp:
@L1:    lda  FIRST_RIDX
        clc
        adc  #RASTERCNT-4   ;4
        cmp  NEXT_RIDX
        bne  @L1
        ldx  FIRSTLINE
        inx
        cpx  #24
        bcc  @L0
        ldx  #0
@L0:    txa
        jsr  setFirstLine
        bit  DISPMODE
        bvc  @L2
        bpl  @L2
        jsr  SCRROUTINE2
@L2:    jmp  SCRROUTINE1


cursorDisable:
        ldy  #0
        sty  ENABLECRSR     ;cursor off
        rts


cursorHide:
        lda  #0
        sta  ENABLECRSR     ;cursor off
        ldy  #6
        lda  CRSRSAVE
        sta  (GFXADR),y
        iny
        lda  CRSRSAVE+1
        sta  (GFXADR),y
        rts


home:   ldx  #0
        ldy  #0

gotoxy: cpx  #80
        bcs  @L0
        cpy  #24
        bcs  @L0
        stx  POSX
        sty  POSY
        jmp  cursorMove
@L0:    rts


gotox:  tax
        ldy  POSY
        jmp  gotoxy


gotoy:  tay
        ldx  POSX
        jmp  gotoxy


printCR:
        lda  #0
        sta  POSX
        jmp  cursorMove


printReturn:
        lda  #0
        sta  POSX

printLF:
        lda  POSY
        cmp  #23
        bcc  @L1
        lda  ENABLECRSR
        beq  @L0
        jsr  cursorHide
@L0:    jsr  scrollUp
        jmp  cursorMove
@L1:    inc  POSY

cursorMove:                 ;move cursor to new position in POSX/POSY
        lda  ENABLECRSR
        beq  @L1
        jsr  cursorHide
@L1:    jmp  setNewCursor


putch:  ;print a character
        ;This routine must be fast!
        ;We use "inline code" to speed up this routine.
        cmp  #32
        bcs  @L1
        ;control characters?
        cmp  #10
        beq  printLF
        cmp  #13
        beq  printCR
        lda  #'0'
@L1:    ;jsr  cursorDisable
        ;inline:
         ldy  #0
         sty  ENABLECRSR     ;cursor off
        sta  LASTPRINTED
        jsr  outputCharacter
        lda  POSX
        cmp  #79
        bcs  printReturn
        inc  POSX
    lsr
    bcs  setNewCursor
    ;fast version of "calcCharGfxAddr"
    lda  #8
    sta  GFXNIB
    lda  #$F0
    sta  GFXNMASK
    jmp  cursorEnable

setNewCursor:
        jsr  calcCharGfxAddr

cursorEnable:
        lda  #0
        sta  CRSRCLKDIV
        ldy  #6
        lda  (GFXADR),y
        sta  CRSRSAVE
        iny
        lda  (GFXADR),y
        sta  CRSRSAVE+1
        sty  ENABLECRSR      ;cursor on
        rts


printHex8:
        pha
        lsr
        lsr
        lsr
        lsr
        jsr  @L1
        pla
@L1:    and  #15
        tay
        lda  hextab,y
        jmp  putch


clearScreen:
        ldy  #240
@L2:    lda  #255
        ldx  #30
@L1:    sta  1063,y
        sta  1303,y
        sta  1543,y
        sta  1783,y
        dey
        dex
        bne  @L1
        tya
        tay
        bne  @L2
        jmp  home


clearLine:
        ldy  POSY
        lda  crtAdrl,y
        sta  PTR1
        lda  crtAdrh,y
        sta  PTR1+1
        lda  POSX
        lsr
        bcc  @L2
        pha
        jsr  cursorDisable
        ldy  #7
@L3:    lda  (GFXADR),y
        and  #$F0
        sta  (GFXADR),y
        dey
        bpl  @L3
        jsr  cursorEnable
        pla
@L2:    tay
        lda  #255
@L1:    sta  (PTR1),y
        iny
        cpy  #40
        bcc  @L1
        rts


doCursor:  ;let the cursor blink
        lda  ENABLECRSR
        beq  @L1
        lda  CRSRCLKDIV
        bne  @L1
        lda  #14
        sta  CRSRCLKDIV
        ;toggle the cursor
        ldy  #0
        lda  GFXCRTCODE
        cmp  (GFXCRTADR),y
        beq  @L0
        bit  DISPMODE
        bvc  @L2
        bpl  @L2
        lda  GFXCRTADR
        sta  @CPTR+1
        lda  GFXCRTADR+1
        clc
        adc  #$D4
        sta  @CPTR+2
        lda  FGCOLOR
@CPTR:  sta  $D800
@L2:    tya
        sta  (GFXADR),y
        iny
        sta  (GFXADR),y
        iny
        sta  (GFXADR),y
        iny
        sta  (GFXADR),y
        iny
        sta  (GFXADR),y
        iny
        sta  (GFXADR),y
        sta  CRSRSAVE
        sta  CRSRSAVE+1
        lda  GFXNMASK
        eor  #$FF
        iny
        sta  (GFXADR),y
        iny
        sta  (GFXADR),y
        ldy  #0
        lda  GFXCRTCODE
        sta  (GFXCRTADR),y
        rts
@L0:    ldy  #6
        lda  (GFXADR),y
        eor  GFXNMASK
        eor  #$FF
        sta  (GFXADR),y
        iny
        lda  (GFXADR),y
        eor  GFXNMASK
        eor  #$ff
        sta  (GFXADR),y
@L1:    rts


printByte:
        pha
        jsr  printHex8
        lda  #32
        jsr  putch
        pla
        rts


printLine:
        jsr  printReturn
        ldx  #80
@L1:    txa
        pha
        lda  #95
        jsr  putch
        pla
        tax
        dex
        bne  @L1
        rts


switchToHexMode:
        lda  DISPMODE
        and  #$7F
        sta  DISPMODE
        jsr  setHexMonoCol
        lda  LASTPRINTED
        beq  swPrHd
        jsr  printLine
swPrHd: lda  #0
        sta  LASTPRINTED
        sta  CURHEXROW
        jmp  printHeadline


switchToTextMode:
        lda  LASTPRINTED
        beq  @L0
        jsr  printLine
@L0:    jsr  setHexColCol
        lda  DISPMODE
        ora  #$80
        sta  DISPMODE
        jmp  swPrHd


outputHex:
        pha
        jsr  printHex8
        lda  CURHEXROW
        clc
        adc  #60
        jsr  gotox
        pla
        cmp  #32
        bcs  @L1
        lda  #'.'
@L1:    inc  REVERSE
        jsr  putch
        dec  REVERSE
        lda  POSX
        beq  @L2
        inc  CURHEXROW
        lda  CURHEXROW
        asl
        adc  CURHEXROW
        jmp  gotox
@L2:    sta  CURHEXROW
        rts


initColorMem:
        ldy  #0
@loop:  sta  $d828,y
        sta  $d900,y
        sta  $da00,y
        sta  $db00,y
        dey
        bne  @loop
        rts


setMonochromeMode:
        lda  #MONOFGCOL
        sta  FGCOLOR
        jsr  initColorMem
        lda  #MONOFRAMECOL
        sta  $d020
        lda  #MONOBGCOL
        sta  $d021
        sta  BGCOLOR
        lda  DISPMODE
        and  #$BF
        sta  DISPMODE
        jmp  printHeadline

setColorMode:
        lda  #15
        sta  FGCOLOR
        ldx  #0
        stx  BGCOLOR
        lda  DISPMODE
        ora  #$40
        sta  DISPMODE
        and  #$80
        beq  @L1
        stx  $d021
        lda  #11
        sta  $d020
        jmp  printHeadline
@L1:    jsr  setHexMonoCol
        jmp  printHeadline

setHexMonoCol:
        lda  #MONOFRAMECOL
        sta  $d020
        lda  #MONOBGCOL
        sta  $d021
        lda  #MONOFGCOL
        jmp  initColorMem

setHexColCol:
        bit  DISPMODE
        bvc  @L1  
        lda  BGCOLOR
        sta  $d021
        lda  #11
        sta  $d020
@L1:    rts



;------------------------------------------------------------------------------
; Scan Keyboard Function (a big part is copied from the kernel ROM)


scanKeyboard:
        LDA #$00
        STA $028D
        LDY #$40
        STY $CB
        STA $DC00
        LDX $DC01
        CPX #$FF
        BNE @L12
        JMP @L15
@L12:   TAY
        LDA #$81
        STA $F5
        LDA #$EB
        STA $F6
        LDA #$FE
        STA $DC00
@L1:    LDX #$08
        PHA
@L2:    LDA $DC01
        CMP $DC01
        BNE @L2
@L3:    LSR
        BCS @L6
        PHA
        LDA ($F5),Y
        CMP #$05
        BCS @L4
        CMP #$03
        BEQ @L4
        ORA $028D
        STA $028D   ;flags for shift(bit0), C=(bit1), CTRL(bit2)
        BPL @L5
@L4:    STY $CB
@L5:    PLA
@L6:    INY
        CPY #$41
        BCS @L7
        DEX
        BNE @L3
        SEC
        PLA
        ROL
        STA $DC00
        BNE @L1
@L7:    PLA
        LDA $028D
        CMP #$03
        BNE @L8
        CMP $028E
        BEQ @L18
        LDA $0291
        BMI @L10
        LDA $D018
        EOR #$02
        STA $D018
        JMP @L10
        ; select keyboard table
@L8:    ASL
        CMP #$08
        BCC @L9
        LDA #$06
@L9:    TAX
        LDA $EB79,X
        STA $F5
        LDA $EB7A,X
        STA $F6
@L10:   LDY $CB
        LDA ($F5),Y
        TAX
        CPY $C5
        BEQ @L11
        LDY #$08   ;set initial wait
        STY $028C
        BNE @L15
@L11:   AND #$7F
        BIT $028A
        BMI @L13
        BVS @L16
        CMP #$7F
        BEQ @L15
@L13:   LDY $028C
        BEQ @L14
        DEC $028C
        BNE @L16
@L14:   DEC $028B
        BNE @L16
        LDY #$02   ;set key auto repeat time
        STY $028B
        LDY $C6
        DEY
        BPL @L16
@L15:   LDY $CB
        STY $C5
        LDY $028D
        STY $028E
        CPX #$FF
@L18:   BEQ @L16
        ;change meaning of keys
        lda  $028D
        cpx  #95
        bne  @L21
        ldx  #27    ;left arrow will change to ESC
        and  #2
        beq  @L17
        bne  startBasic  ;c= + ESC will start Basic
@L21:   cpx  #'-'
        bne  @L19
        and  #1
        bne  @L20
@L19:   cpx  #61
        bne  @L17
        lda  $028D
        and  #1
        beq  @L17
@L20:   ldx  #95  ;shift + = -> _
@L17:   ;put key (in X) to buffer
        TXA
        LDX $C6
        CPX $0289
        BCS @L16
        STA $0277,X
        INX
        STX $C6
@L16:   LDA #$7F
        STA $DC00
        RTS


startBasic:
        sei
        ldx  #$FF
        txs
        cld
        inx
        stx  $D016
        jsr  $FDA3
        jsr  $FD50
        jsr  $FD8C
        jsr  $FD15
        jsr  $FF5B
        cli
        jmp  ($A000)
        



;------------------------------------------------------------------------------
; Buffer Functions


initUartBuffer:
        lda  #<RXBUFBASE
        sta  RXB_INPTR
        sta  RXB_OUTPTR
        lda  #>RXBUFBASE
        sta  RXB_INPTR+1
        sta  RXB_OUTPTR+1
        ldy  #0
        sty  TXB_INPTR
        sty  TXB_OUTPTR
        lda  #0
@L0:    sta  RXBFILLTAB,y
        iny
        bne  @L0
        rts


putToTxBuf:
        ldy  TXB_INPTR
        sta  TXBUFFER,y
        inc  TXB_INPTR
        rts

hasTxbufSpace:
        lda  TXB_INPTR
        sec
        sbc  TXB_OUTPTR
        bcs  @L1
        adc  #<TXBUFSIZE
@L1:    cmp  #TXBUFSIZE/2
        rts   ;returns with C=0 when buffer has still space


getFromRxBuf:
        ;read byte from buffer, return byte in Accu
        ;returns with zeroflag set when buffer is empty
        
        ldx  RXB_OUTPTR+1
        cpx  RXB_INPTR+1
        beq  @L0
        lda  RXBFILLTAB,x
        beq  @L2
        tax
        ldy  #0
        lda  (RXB_OUTPTR),y
        pha
        inc  RXB_OUTPTR
        cpx  RXB_OUTPTR
        bne  @L1
        sty  RXB_OUTPTR
        ldx  RXB_OUTPTR+1
        jsr  @L4
@L1:    pla
        beq  getFromRxBuf
@L0:    rts        

@L2:    jsr  @L4
        jmp  getFromRxBuf
@L4:    inx
        cpx  #UPPER_RXBUF
        bcc  @L3
        ldx  #>RXBUFBASE
@L3:    stx  RXB_OUTPTR+1
        rts
        

.if ansitest

processTestData:
        ldx  #<bufinitstart
        ldy  #>bufinitstart
        stx  @src+1
        sty  @src+2
@loop:  lda  @src+1
        cmp  #<bufinitend
        bne  @src
        lda  @src+2
        cmp  #>bufinitend
        beq  @L0
@src:   lda  $4100
        jsr  processIncomingData
        jsr  mainServices
        inc  @src+1
        bne  @loop
        inc  @src+2
        jmp  @loop        
@L0:    rts

.endif




;------------------------------------------------------------------------------
; UART Functions



handshakeON:  ;switch peer transmitter on
        lda  HANDSHAKE
        cmp  #2
        beq  @L1
        ;hardware handshake
        lda  UART_MCR
        ora  #$02       ;set /RTS to low-level
        sta  UART_MCR
@L0:    rts
@L1:    ;software handshake
        lda  #XON
        sta  NEXTHCTOSND
        rts


uartoutput:
        ;first step: try to send a handshake character
        ldy  NEXTHCTOSND
        cpy  LASTHCSENT
        beq  @L2   ;no handshake character to transmit
        lda  UART_LSR
        and  #$20
        beq  @L0   ;transmitter not ready
        sty  LASTHCSENT
        lda  HANDSHAKE
        cmp  #2
        bne  @L2   ;no software handshake
        sty  UART_THR  ;send out handshake character

@L2:    ldy  TXB_OUTPTR
        cpy  TXB_INPTR
        beq  @L0   ;nothing to tranmsit
        lda  HANDSHAKE
        cmp  #1
        bne  @L1   ;no hardware handshake
.if swtest
        lda  #$10
.else
        lda  UART_MSR
.endif
        and  #$10
        beq  @L0   ;handshake: transmitter off
@L1:
.if swtest
        lda #$20
.else
        lda  UART_LSR
.endif
        and  #$20
        beq  @L0   ;transmitter not ready
        inc  TXB_OUTPTR
        lda  TXBUFFER,y
        sta  UART_THR
.if swtest
        jsr  processIncomingData
.endif
        jmp  uartoutput
@L0:    rts


sendXON:
        lda  HANDSHAKE
        cmp  #2
        bne  @L0
        ldy  #XON
        sty  NEXTHCTOSND
        sta  LASTHCSENT
@L0:    rts


setupUART:
        ;set baud rate
        ldx  #0
        stx  ENABLEIRQ
        stx  UART_IER
        lda  #$80
        sta  UART_LCR
        ldy  BAUDRATE
        lda  brDivisors,y
        sta  UART_DLL
        stx  UART_DLM
        ;set data/stop/parity-bits
        lda  DATABITS
        sec
        sbc  #5
        and  #3
        sta  TMP1
        lda  STOPBITS
        asl
        and  #$04
        ora  TMP1
        ldy  PARITY
        beq  @L1
        ora  #$08
        dey
        beq  @L1
        ora  #$10
@L1:    sta  UART_LCR
        ;setup other registers
        stx  UART_IER
        ldy  #16
@L2:    lda  UART_IIR
        lda  UART_MSR
        dey
        bne  @L2
        lda  #$01
        sta  UART_FCR
        ora  #$06
        sta  UART_FCR   ;enable FIFO, set RX-threshold to 1
        lda  #MCRDEFAULT
        ora  #$02       ;set /RTS to low-level
        sta  UART_MCR
@L3:    lda  UART_RBR
        lda  UART_LSR
        lsr
        bcs  @L3
        inc  ENABLEIRQ  ;switch on UART interrupt
        rts


getBaudText:   ;returns y as index to baudRates strings
        ldy  #0
        ldx  BAUDRATE
        beq  @L1
@L2:    iny
        lda  baudRates,y
        bne  @L2
        iny
        dex
        bne  @L2
@L1:    rts



printHeadline:
        ldy  #0
@L1:    bit  DISPMODE
        bpl  @L8
        bvc  @L10
        lda  headlineT,y
        ldx  #4   ;color
        jmp  @L9
@L10:   lda  headlineT,y
        ldx  $d020
        jmp  @L9
@L8:    lda  headlineH,y
        ldx  $d020
@L9:    ora  #$80
        sta  $0400,y
        txa
        sta  $d800,y
        iny
        cpy  #40
        bcc  @L1
        lda  INITFLAG
        beq  @L5
        jsr  getBaudText
        ldx  #0
@L6:    lda  baudRates,y
        beq  @L7
        iny
        inx
        bne  @L6
@L7:    stx  TMP1
        lda  #6
        sec
        sbc  TMP1
        pha
        jsr  getBaudText
        pla
        tax
@L2:    lda  baudRates,y
        beq  @L3
        ora  #$80
        sta  $0400+22,x
        inx
        iny
        bne  @L2
@L3:    lda  #','+128
        sta  $0400+28
        sta  $0400+30
        sta  $0400+32
        sta  $0400+34
        lda  DATABITS
        ora  #$B0
        sta  $0400+29
        lda  STOPBITS
        ora  #$B0
        sta  $0400+31
        ldy  PARITY
        lda  parities,y
        ora  #$80
        sta  $0400+33
        ldy  HANDSHAKE
        lda  handshakes,y
        ora  #$80
        sta  $0400+35
        lda  CHARECHO
        beq  @L4
        lda  #'+'+128
        sta  $0400+37
        lda  #'e'+128
        sta  $0400+38
@L4:    lda  APPENDLF
        beq  @L5
        lda  #'+'+128
        sta  $0400+37
        lda  #'l'+128
        sta  $0400+39
@L5:    rts




;------------------------------------------------------------------------------
; I2C Functions


i2cWrSDA:
        lda  I2COUTREG
        bcc  @L1
        and  #255-I2COUT_SDA
        bcs  @L2
@L1:    ora  #I2COUT_SDA
@L2:    sta  I2COUTREG
        rts

i2cRdSDA:
        lda  I2CINREG
        eor  #I2CIN_SDA
        rol
        rts

i2cWrSCL:
        lda  I2COUTREG
        bcc  @L1
        and  #255-I2COUT_SCL
        bcs  @L2
@L1:    ora  #I2COUT_SCL
@L2:    sta  I2COUTREG
        rts

i2cRdSCL:
        lda  I2CINREG
        eor  #I2CIN_SCL
        rol
        rol
        rol
        rts

i2cStop:
        clc
        jsr  i2cWrSDA
        sec
        jsr  i2cWrSCL
        jmp  i2cWrSDA

i2cStart:
        sec
        jsr  i2cWrSCL
        jsr  i2cWrSDA
        clc
        jsr  i2cWrSDA
        jmp  i2cWrSCL

i2cSendByte:  ;Send out 1 databyte (in Accu). Returns with C=0 on success.
        ldx  #8
@L1:    rol
        pha
        jsr  i2cWrSDA
        sec
        jsr  i2cWrSCL
        ldy  #10
@L2:    dey
        bne  @L2
        clc
        jsr  i2cWrSCL
        pla
        dex
        bne  @L1
        sec
        jsr  i2cWrSDA
        jsr  i2cWrSCL
        ldx  #50
@L5:    jsr  i2cRdSCL
        bcs  @L4
        dex
        bne  @L5
        sec
        rts
@L4:    jsr  i2cRdSDA
        php
        clc
        jsr  i2cWrSCL
        jsr  i2cWrSDA
        plp
        rts

i2cRecvByte:  ;Receive 1 byte. Returns databyte in Accu and C=0 on success.
              ;input: C=0: achknowledge the byte, C=1: do not ACK the byte
        php
        sec
        jsr  i2cWrSDA
        ldx  #8
        pha
@L1:    sec
        jsr  i2cWrSCL
        jsr  i2cRdSDA
        pla
        rol
        pha
        clc
        jsr  i2cWrSCL
        ldy  #3
@L2:    dey
        bne  @L2
        dex
        bne  @L1
        pla
        plp
        pha
        jsr  i2cWrSDA
        sec
        jsr  i2cWrSCL
        ldy  #9
@L3:    dey
        bne  @L3
        clc
        jsr  i2cWrSCL
        pla
        clc
        rts

i2cRead:     ;read 1 byte. Input: X = memory address, Output: A = byte, C=0 on success
        txa
        pha
        jsr  i2cStart
        lda  #EE24C16ADR
        jsr  i2cSendByte
        pla
        bcs  @L1
        jsr  i2cSendByte
        bcs  @L1
        jsr  i2cStart
        lda  #EE24C16ADR+1
        jsr  i2cSendByte
        sec
        jsr  i2cRecvByte
        php
        pha
        jsr  i2cStop
        pla
        plp
        rts
@L1:    ;error
        jsr  i2cStop
        sec
        rts

i2cWrite:   ;write 1 byte. Input: X = memory address, A = byte, Output: C=0 on success
        pha
        txa
        pha
        jsr  i2cStart
        lda  #EE24C16ADR
        jsr  i2cSendByte
        pla
        bcs  @L1
        jsr  i2cSendByte
        bcs  @L1
        pla
        jsr  i2cSendByte
        php
        jsr  i2cStop
        jsr  wait6ms  ;wait until EEPROM-device has written the byte to flash memory
        plp
        rts
@L1:    ;error
        pla
        jsr  i2cStop
        sec
        rts

wait6ms:
        ldy  #6  ;6ms
@L2:    ldx  #249  ;2
@L1:    dex        ;2
        bne  @L1   ;2
        dey        ;2
        bne  @L2   ;2
        rts


i2cReadConfig:    ;returns with C=0 on success
        jsr  _i2cReadCfg
        bcc  @L0
        lda  MEMMAGIC
        cmp  #MAGICNUM
        beq  @L0  ;C=1: error
        ;EEPROM not initialized. Initialize now.
        jsr  defaultConfig
        jsr  i2cWriteConfig
        jsr  _i2cReadCfg
@L0:    rts


_i2cReadCfg:    ;returns with C=0 on success
        ldx  #0
@L1:    txa
        pha
        jsr  i2cRead
        tay
        pla
        tax
        tya
        bcs  @L2  ;error
        sta  CFG_START,x
        inx
        cpx  #CFG_SIZE
        bcc  @L1
        lda  MEMMAGIC
        cmp  #MAGICNUM
        bne  @L2  ;error
        jsr  calcChecksum
        cmp  CHECKSUM
        bne  @L2  ;error
        jsr  checkConfig
        clc
        rts
@L2:    sec
        rts


i2cWriteConfig:   ;returns with C=0 on success
        lda  #MAGICNUM
        sta  MEMMAGIC
        jsr  calcChecksum
        sta  CHECKSUM
        ldx  #0
@L1:    lda  CFG_START,x
        tay
        txa
        pha
        tya
        jsr  i2cWrite
        pla
        bcs  @L2  ;error
        tax
        inx
        cpx  #CFG_SIZE
        bcc  @L1
        clc
@L2:    rts
        

checkEEPROMwrite:
        lda  WRITETIMER
        beq  @L0
        dec  WRITETIMER
        bne  @L0
        jsr  sendXON ;in case XON is now selected
        jsr  i2cWriteConfig
        bcc  @L0
        lda  EEPROMERR
        bne  @L0
        ldx  #<eepWrErr
        ldy  #>eepWrErr
        jsr  printString
@L0:    rts        


defaultConfig:
        lda  #DEF_BAUD
        sta  BAUDRATE
        lda  #DEF_DATA
        sta  DATABITS
        lda  #DEF_STOP
        sta  STOPBITS
        lda  #DEF_PARITY
        sta  PARITY
        lda  #DEF_HAND
        sta  HANDSHAKE
        lda  #DEF_LF
        sta  APPENDLF
        lda  #DEF_ECHO
        sta  CHARECHO
        lda  #$80
        sta  DISPLAYMODE
        rts


checkConfig:
        lda  BAUDRATE
        cmp  #MAXBAUDRATES+1
        bcc  @L1
        lda  #DEF_BAUD
        sta  BAUDRATE
@L1:    lda  DATABITS
        cmp  #5
        bcc  @L2
        cmp  #9
        bcc  @L3
@L2:    lda  #DEF_DATA
        sta  DATABITS
@L3:    lda  STOPBITS
        cmp  #3
        bcc  @L4
        lda  #DEF_STOP
        sta  STOPBITS
@L4:    lda  HANDSHAKE
        cmp  #3
        bcc  @L5
        lda  #DEF_HAND
        sta  HANDSHAKE
@L5:    lda  APPENDLF
        cmp  #2
        bcc  @L6
        lda  #DEF_LF
        sta  APPENDLF
@L6:    lda  CHARECHO
        cmp  #2
        bcc  @L7
        lda  #DEF_ECHO
        sta  CHARECHO
@L7:    lda  PARITY
        cmp  #3
        bcc  @L8
        lda  #DEF_PARITY
        sta  PARITY
@L8:    rts



calcChecksum:
        ldx  #0
        lda  #$FF
@L1:    eor  CFG_START,x
        asl
        bcc  @L2
        eor  #$21
@L2:    inx
        cpx  #CFG_SIZE-1
        bcc  @L1
        rts

